// File OGLCtrl.h

#ifndef OGLCTRL_H
#define OGLCTRL_H

#if _MSC_VER >= 1000
#pragma once
#endif 

#include "gl\gl.h"     // Include the standard OpenGL headers
#include "gl\glu.h"    // Include the OpenGL utility library
#include "gl\glaux.h"  // Add in the auxiliary library?

// To link you will need opengl32.lib, glu32.lib and perhaps glaux.lib
// You will need opengl32.dll and glu32.dll to run the application too

//*****************************************************************************
// OpenGL reusable window classes to encapsulate a window with an OpenGL graph
// By G.Bavestrelli,                             Copyright Techint S.p.A. 1998
//*****************************************************************************

//=============================================================================
// Class COpenGLCtrl:To be used if you have only one OpenGL control in your app
//=============================================================================

class COpenGLCtrl : public CWnd
{
    DECLARE_DYNAMIC(COpenGLCtrl);   

    CRect m_ClipBox;

  protected:

    HGLRC m_hRC;  // Holds the rendering context
    HDC   m_hDC;  // Holds the device context

  public:  

    COpenGLCtrl(); 
   ~COpenGLCtrl();      

    // Register "MyOpenGL" window class
    static BOOL RegisterOpenGLWindowClass();

 protected:

    // Initialization and uninitialization 
    // You probably won't need to redefine them
    virtual void InitializeOpenGLWindow();
    virtual void UninitializeOpenGLWindow();
 
    // Set the Pixel format
    // Redefine CustomizePixelFormat to change some settings in the PIXELFORMATDESCRIPTOR,
    // or SetupPixelFormat to take into your hands the whole process.
    virtual void SetupPixelFormat(HDC hDC);
    virtual void CustomizePixelFormat(PIXELFORMATDESCRIPTOR * pfd){}

    // These functions are called to handle WM_PAINT and WM_SIZE messages
    // They implement the protocol further below
    virtual void RedrawOpenGLWindow();
    virtual void ResizeOpenGLWindow(int cx, int cy);
  
    // What follows is a protocol defined in the above functions
    // These are the functions you should redefine to draw your scene
    virtual void OnCreateRC();
    virtual void OnViewport( int cx, int cy );
    virtual void OnProjection(GLdouble AspectRatio);
    virtual void OnViewing();
    virtual void OnRender(); 
    virtual void OnDestroyRC();

    // These are here only for compatibility, they do nothing here
    // They are only used by derived classes
    virtual void BeginOpenGLDrawing(){}
    virtual void EndOpenGLDrawing(){}

    // If you need the clip region, here it is
    const CRect & GetClipRect() const { return m_ClipBox; }

// Overrides
    // ClassWizard generated virtual function overrides
    //{{AFX_VIRTUAL(COpenGLCtrl)
    protected:
    virtual BOOL PreCreateWindow(CREATESTRUCT& cs);
    virtual void PreSubclassWindow();
    //}}AFX_VIRTUAL


    // Generated message map functions
protected:
    //{{AFX_MSG(COpenGLCtrl)
    afx_msg void OnDestroy();
    afx_msg void OnSize(UINT nType, int cx, int cy);
    afx_msg void OnPaint();
    afx_msg BOOL OnEraseBkgnd(CDC* pDC);
    //}}AFX_MSG
    DECLARE_MESSAGE_MAP()
};

//=============================================================================
// Class COpenGLCtrls : To be used if you have many OpenGL controls in your app
//=============================================================================

class COpenGLCtrls : public COpenGLCtrl
{
    DECLARE_DYNAMIC(COpenGLCtrls);   
    
    int m_CountBeginEnd; // Count calls to BeginDrawing and EndDrawing

  public:

    COpenGLCtrls();      
   ~COpenGLCtrls();      

    virtual void InitializeOpenGLWindow();
    virtual void UninitializeOpenGLWindow();

    virtual void RedrawOpenGLWindow();
    virtual void ResizeOpenGLWindow(int cx, int cy);

    virtual void BeginOpenGLDrawing();
    virtual void EndOpenGLDrawing();
};


// Function that returns TRUE if OpenGL reported an error
BOOL GetOpenGLError();

//{{AFX_INSERT_LOCATION}}

#endif // OGLCTRL_H
